unit mfc;   {Version 2.0}

{ Programming unit for the IBM Music Feature Card and Roland MPU-401
  compatible MIDI cards.  (See unit FB01 for sound module / MIDI functions.)

  Copyright (C) 1989, 1990,  Kevin Weiner

 }

INTERFACE

const

{ Node status parameter offsets
  $00-$0d also correspond to equivalent configration data offsets }

  node_name     = $00;
  node_combine  = $08;
  node_lfospeed = $09;
  node_amd      = $0a;
  node_pmd      = $0b;
  node_lfowave  = $0c;
  node_recmode  = $0d;
                          { Offsets in nstat array }
  node_number   = $20;      stat_number   = 0;
  node_protect  = $21;      stat_protect  = 1;
  node_config   = $22;      stat_config   = 2;
  node_tune     = $23;      stat_tune     = 3;
  node_output   = $24;      stat_output   = 4;
  node_chain    = $25;      stat_chain    = 5;

  maxpacket = 512;                     {Max packet size supported}

  iodefault = 500;                     {Default I/O time out (ms)}

type

{ MIDI paths through MFC }
  paths = (midiin_system, system_midiout, midiin_sp, system_sp,
           midiin_midiout);

{ Array of 5, 8-bit path enable indicators:  000edcba
      a = Note on/off
      b = After touch / pitch bend
      c = Control/Program change
      d = System Exclusive/Common
      e = System real-time            (Normally, all are set to 1 to enable) }
  pathdata = array [paths] of byte;

{ Array of master node parameters }
  nodedata = array [0..5] of byte;

var
  node, currentmfc: integer;           {Current node and current MFC}
  mfcmode, mpumode,                    {MFC or MPU-401 in use}
  cardpresent,                         {MFC or MPU available}
  mfc0, mfc1: boolean;                 {MFC 0 active; 1 active}
  pstat: array [0..1] of pathdata;     {Paths set for local MFC's}
  nstat: array [0..1] of nodedata;     {Node parameters for local MFC's}
  iolimit: integer;                    {I/O time out (ms)}
  packetbuffer: array[1..maxpacket] of byte;
  packetsize: integer;
  packeterror: boolean;
  lastdata: integer;                   {Last data byte read}
  realtime: boolean;                   {MIDI real-time messages passed}
  midisent: boolean;                   {MIDI data was written}
  mfctimermode: boolean;               {Enable MFC timing}
  mdebug: boolean;                     {Debug mode flag}
  waitack: boolean;                    {Require ack after dumps}
  intmode: boolean;                    {Use interrupt mode}
  irq: integer;                        {IRQ number selected}

  counter: integer;  {****}

{ MFC user routines ---------------------------------------------------------}

function inputready: boolean;          {Test if input is waiting}

procedure getbyte(var data: integer);  {Read a data byte from the interface}

procedure getdata(var x: integer);     {Get byte, skipping realtime}

procedure setdata;                     {Set data mode for interface writes}

procedure setcmd;                      {Set command mode for interface writes}

procedure putbyte(data: integer);      {Write a byte to the interface}

procedure waitbyte(data: integer);     {Wait for specified byte or timeout}

procedure clearinput;                  {Clear pending MIDI input}

procedure path(pathid: paths; status: boolean);  {Set selected path on/off
                                                  (MFC only)}

procedure getstatus;                   {Read node status to nstat array (MFC)}

procedure putstatus;                   {Write node parameters (MFC)}

procedure sendnoteon(channel, note, vel: integer);   {Play note on channel}

procedure controlchange(channel, control, value: integer);
                                       {Issue control change on channel}

procedure quiet;                       {Stop all notes in progress}

procedure alloff(channel: integer);    {Stop notes playing on a channel}

procedure selectvoice(channel, voiceno: integer);   {Select voice on channel}

procedure selectmfc(card: integer);    {Select current MFC - 0/1}

procedure reboot;                      {Reset (both MFC's) to power-on state}

procedure testcard;                    {Diagnostic test for MFC}

procedure settimer(n: integer);        {Set MFC countdown timer}

function timeout: boolean;             {Test MFC timer <= 0}

procedure reenableint;


IMPLEMENTATION

uses crt, dos, queue;

const
  mfcbase0 = $2a20;                    {Base port address of first MFC}
  mfcbase1 = $2a30;                    {Second MFC}
  mpubase0 = $330;                     {Default MPU port address}

{MFC port offsets}
  piu0  = 0;                           {Parallel interface unit - port 0}
  piu1  = 1;                           {Port 1}
  piu2  = 2;                           {Port 2}
  pcr   = 3;                           {PIU command register}
  cntr0 = 4;                           {Counter register 0}
  cntr1 = 5;                           {Counter register 1}
  cntr2 = 6;                           {Counter register 2}
  tcwr  = 7;                           {Timer control word register}
  tcr   = 8;                           {Total control register}
  tsr   = 12;                          {Total status register}

{MFC flag bits}
  txrdy = 2;                           {MFC ready for write}
  rxrdy = $20;                         {Data ready to read}
  exr8  = $80;                         {8th bit - command read}
  ext8  = $10;                         {8th bit - command write}
  ibe   = $80;                         {Interrupt enable}
  tmask = $40;                         {Mask enable}
  cmd_bit = $100;                      {9th bit set in data for command}

{PIU commands}
  initpiu = $bc;                       {Initialize}
  setread = $09;                       {Set read interrupt enable}
  resread = $04;                       {Reset rie}


{MFC timer options - sum for control word commands}
  timer0 = $00;                        {Timer select}
  timer1 = $40;
  timer2 = $80;
  latch = $00;                         {Latch timer value to register}
  lsb_only = $10;                      {Read/write least significant byte}
  msb_only = $20;                      {Read/write most siginificant byte}
  lsb_msb = $30;                       {Read/write least sig., then most}
  mode2 = $04;                         {Timer mode 2}
  mode3 = $06;                         {Timer mode 3}

{MFC commands}
  cmd_set_path = $1e2;                 {Set path command}
  cmd_reboot = $1e5;                   {Reboot command}
  cmd_getpathstatus = $1d2;            {Get path status command}
  cmd_setnodestatus = $1e3;            {Set node status command}
  cmd_getnodestatus = $1d3;            {Get node status command}
  cmd_getcardmode = $1d0;              {Get card mode (music/thru)}

{MPU constants}
  mpudata = 0;                         {Data port offset}
  mpustat = 1;                         {Status port offset}
  drs = $80;                           {Data ready to send}
  drr = $40;                           {Data ready to receive}
  mpuack = $fe;                        {Command acknowledge}
  cmd_uart = $3f;                      {Basic serial i/o mode}
  cmd_reset = $ff;                     {Reset MPU}


{Misc. constants}
  retries = 2;                         {Retry limit on packet errors}
  hexdigits: string[16] = '0123456789ABCDEF';

type

{ Packet type selection: a = 8-bit data,  b = 7-bit data }
  packettype = (a_packet, b_packet);

var
  mfcbase: integer;                    {Current MFC base port address}
  mpubase: integer;                    {MPU base port address}
  commandmode: boolean;                {Data to be written in command mode}
  istat: array [0..1] of byte;         {Starting PIU2 status (MFC)}
  selectmpu: boolean;                  {Force use of MPU}
  exitsave: pointer;                   {Saved program exit address}
  lastack: boolean;                    {Last dump ack result}

  timeoutset: boolean;                 {MFC counter timed out}
  timerhigh, timerlow: integer;        {Zero point for timer in progress}
  starthigh, startlow: integer;        {Initial timer value}
  timercheck: integer;                 {Non-functioning timer safety limit}

  starttime, checktime: longint;       {Starting time and limit for DOS clock}

  intinstalled: boolean;               {interrupt routine installed}
  intenabled: boolean;                 {Interrupt enabled}
  intselected: boolean;                {-i option used}

  tcrbyte: byte;                       {TCR contents}
  intvec: byte;                        {Interrupt vector number}
  irqmask: byte;                       {IRQ mask for PIC}
  savevec: pointer;                    {Saved interrupt vector}


function inport(x: integer): byte;
  inline($5a/$eb/$00/$ec);

function hex(n: integer): string;

  { Convert integer to hex string }

  var
    s: string;

  begin
    if n < 0 then
      hex := '-1'
    else
      begin
        s := '';
        repeat
          s := hexdigits[(n mod 16)+1] + s;
          n := n div 16;
        until n = 0;
        hex := s;
      end;
  end;

procedure enableint;  forward;

procedure disableint; forward;

function mfcready: boolean;

  {Returns true if the MFC has data to be read}

  begin
    mfcready := (inport(mfcbase+piu2) and rxrdy) <> 0
  end;

function mpuready: boolean;

  {Returns true if the MPU has data to be read}

  begin
    mpuready := (inport(mpubase+mpustat) and drs) = 0;
  end;

procedure getmfc(var data: integer);

  {Returns a data byte read from the MFC, or -1 if no data.
   9th bit set on return if command read}

  var
    stat: byte;
    t: boolean;
    n: integer;

  begin
    data := -1;
    if intenabled then                 {Wait for interrupt routine to get data}
      begin
        settimer(iolimit);
        repeat
        until timeout or (queued <> 0);
        if queued > 0 then
          begin
            disableint;
            dequeue(data);
            enableint;
          end;
      end;
    if data < 0 then
      begin
        t := false;
        n := mfcbase + piu2;
        stat := inport(n);            {Read status}
        if (stat and rxrdy) = 0 then
          begin
            settimer(iolimit);             {Wait for data}
            repeat
              stat := inport(n);
              t := timeout;
            until ((stat and rxrdy) <> 0) or t;
          end;
        if not t then
          begin
            n := mfcbase+piu0;
            data := inport(n);        {Check if command read}
            if (stat and exr8) <> 0 then
              data := data + cmd_bit;      {Add in command bit 9}
          end;
      end;
  end;

function time: longint;

  {Return system time in milliseconds}

  var
    h, m, s, hs: word;

  begin
    gettime(h,m,s,hs);
    time := (((h*60+m)*60+s)*100+hs)*10;
  end;

procedure getmpu(var data: integer);

  {Returns a data byte read from the MPU, or -1 if no data.}

  var
    stat: byte;
    t: boolean;

  begin
    data := -1;
    if intenabled then                 {Wait for interrupt routine to get data}
      begin
        settimer(iolimit);
        repeat
        until timeout or (queued <> 0);
        if queued > 0 then
          begin
            disableint;
            dequeue(data);
            enableint;
          end;
      end;
    if data < 0 then
      if (inport(mpubase+mpustat) and drs) = 0 then
        data := inport(mpubase+mpudata)    {Read data}
      else
        begin                            {Wait for data}
          settimer(iolimit);
          repeat
            stat := inport(mpubase+mpustat);
            t := timeout;
          until ((stat and drs) = 0) or t;
          if not t then
            data := inport(mpubase+mpudata);
      end;
  end;

procedure getbyte { var data: integer };

  {Read a data byte from the active interface; returns -1 if none.}

  begin
    if queued > 0 then
      begin
        disableint;
        dequeue(data);
        enableint;
      end
    else if mfcmode then
      getmfc(data)
    else {mpumode}
      getmpu(data);
    lastdata := data;
    if mdebug then write('<',hex(data));
  end;

procedure getdata { var x: integer };

  {Get data byte, skipping realtime data}

  begin
    repeat
      getbyte(x);
    until ((x and $f8) <> $f8) or (x < 0);
  end;

function inputready: boolean;

  var
    x: integer;
    ready, stat, rt: boolean;

  {Returns true if the active interface has data to be read}

  begin
    if mfcmode then
      inputready := mfcready or (queued <> 0)
    else {MPU}
      begin
        ready := mpuready or (queued <> 0);
        if ready then
          begin
            if realtime then
              inputready := true
            else
              begin
                repeat
                  getbyte(x);
                  stat := mpuready;
                  rt := (x and $f8) = $f8;
                until (not rt) or (not mpuready);
                if rt then
                  inputready := false
                else
                  begin
                    inputready := true;
                    disableint;
                    push(x);
                    enableint;
                  end;
              end;
          end
        else
          inputready := false;
      end;
  end;

procedure waitbyte { data: integer } ;

  {Wait for specified data byte to be read or timeout}

  var
    x: integer;

  begin
    if mdebug then write(' wait:',hex(data));
    repeat
      getbyte(x);
    until (x = data) or (x < 0);
  end;

procedure clearinput;

  var
    count, limit, x: integer;

  begin
    if cardpresent then
      begin
        count := 0;
        limit := mmax * 2;
        while inputready and (count < limit) do
          begin
            getbyte(x);
            count := count + 1;
         end;
      end;
  end;

procedure setdata;

  {Set data mode for subsequent interface writes}

  begin
    commandmode := false;
    if mfcmode then
      begin
        port[mfcbase+tcr] := tcrbyte and not ext8;
      end;
  end;

procedure setcmd;

  {Set command mode for subsequent interface writes}

  begin
    commandmode := true;
    if mfcmode then
      begin
        port[mfcbase+tcr] := tcrbyte or ext8;
      end;
  end;

procedure putmfc(data: integer);

  {Write data byte to MFC}

  var
    stat: byte;
    x: integer;

  begin
    stat := inport(mfcbase+piu2);
    if (stat and txrdy) = 0 then
      begin
        settimer(iolimit);
        repeat                            {Wait until MFC ready}
          stat := inport(mfcbase+piu2);
        until ((stat and txrdy) <> 0) or timeout;
      end;
    port[mfcbase+piu1] := byte(data);     {Write data}
  end;

procedure putmpu(data: integer);

  {Write data or command byte to MPU}

  var
    stat: byte;
    t, t1, t2: longint;
    x: integer;

  begin
    if intmode then disableint;
    stat := inport(mpubase+mpustat);
    timeoutset := false;
    if (stat and drr) <> 0 then
      begin                            {Wait until device ready}
        if mpuready then
          begin                        {May need to clear input}
            settimer(iolimit);
            repeat
              getmpu(x);
              if x >= 0 then enqueue(x);
            until (not mpuready) or timeout;
          end;
        settimer(iolimit);
        repeat
          stat := inport(mpubase+mpustat);
        until ((stat and drr) = 0) or timeout;
      end;
    if commandmode then
      begin
        port[mpubase+mpustat] := byte(data);
        delay(100);
        waitbyte(mpuack);
      end
    else
      port[mpubase+mpudata] := byte(data);
    if intmode then enableint;
  end;

procedure putbyte(data: integer);

  {Write data byte to active interface}

  begin
    if mfcmode then
      putmfc(data)
    else
      putmpu(data);
    if mdebug then write('>',hex(data));
    midisent := true;
  end;


procedure ackint;

  inline($ba/$20/$00/$b0/$20/$ee);  {Acknowledge interrupt to PIC}


procedure mfcint;  interrupt;

  var
    x: integer;

  begin
    disableint;
    while mfcready do
      begin
        getmfc(x);
        enqueue(x);
        counter := x;
      end;
    ackint;
    enableint;
  end;

procedure mpuint;  interrupt;

  var x: integer;

  begin
    disableint;
    if mpuready then
      begin
        getmpu(x);
        enqueue(x);
      end;
    enableint;
    ackint;
  end;

procedure enableint;

  begin
    if intmode then
      begin
        if not intinstalled then
          begin
            getintvec(intvec, savevec);
            if mpumode then
              setintvec(intvec, @mpuint)
            else
              begin
                setintvec(intvec, @mfcint);
                tcrbyte := ibe or tmask;
                port[mfcbase+pcr] := initpiu;
                port[mfcbase+tcr] := tcrbyte;
                port[mfcbase+pcr] := setread;
              end;
            intinstalled := true;
          end;
        port[$21] := inport($21) and (not irqmask);
        intenabled := true;
      end;
  end;

procedure disableint;

  begin
    if intenabled then
      begin
        port[$21] := inport($21) or irqmask;
        intenabled := false;
      end;
  end;

procedure reenableint;

  begin
    if intmode then
      begin
        disableint;
        setintvec(intvec, savevec);
        intinstalled := false;
        enableint;
      end;
  end;

procedure sysex;

  {Write 4-byte system exclusive header}

  begin
    putbyte($f0);  putbyte($43);  putbyte($75);  putbyte(node);
  end;

procedure nodedump(source, id: integer);

  {Request dump of current node from specified storage source/id}

  var x: integer;

  begin
    if mdebug then write(' nodedump:',hex(source),':',hex(id));
    sysex;
    putbyte($20);
    putbyte(source);
    putbyte(id);
    putbyte($f7);
  end;

procedure getpacket(p: packettype);

 {Read a packet of the specified type (a_packet=8-bit, b_packet=7-bit).
  Stores packet to packet_buffer with size packet_size.  Packet_error
  indicates read failure if true.}

  var
    x, x1, y, low, size, chksum, testsum: integer;

  begin
    if mdebug then write(' getpacket');
    packeterror := true;
    testsum := 0;
    getbyte(size);                     {get 2-byte packet size}
    getbyte(low);
    size := size * 128 + low;
    if p = a_packet then size := size div 2;
    if mdebug then write(' size=',size);
    if size > 0 then
      begin
        packetsize := 0;
        repeat
          getbyte(x1);
          x := x1;
          testsum := testsum + x;      {Add to checksum}
          if p = a_packet then
            begin
              getbyte(y);              {2nd byte if 8-bit data}
              testsum := testsum + y;
              x := y * 16 + x;
            end;
          packetsize := packetsize + 1;
          if (x < 0) or (x > 256) then x := 0;
          packetbuffer[packetsize] := x;
        until (x1 < 0) or (packetsize = size);
        if mdebug and (x1 < 0) then write(' packet timeout');

        if x1 >= 0 then               {Compare checksums}
          begin
            getbyte(chksum);
            packeterror := ((-testsum) and 127) <> chksum;
            if mdebug and packeterror then write(' checksum error');
          end;
        if mdebug and not packeterror then write(' good packet');
      end;
  end;

function ack: boolean;

  {Wait for status message returned after dump to node.
   Ack true indicates success.}

  var x: integer;

  begin
    if mdebug then write( 'wait ack-');
    waitbyte($f0);
    waitbyte($60+node);
    getbyte(x);
    lastack := x = 2;
    ack := lastack;
    if mdebug then write(lastack);
    getbyte(x);
  end;

procedure setpath(p1,p2,p3,p4,p5: integer);

  {Set all path-enable flags}

  var
    pmask: integer;

  begin
    if mdebug then write(' setpath');
    if realtime then pmask := $1f else pmask := $0f;
    setcmd;
    putbyte(cmd_set_path);
    putbyte(p1 and pmask);  putbyte(p2 and pmask);  putbyte(p3 and pmask);
    putbyte(p4 and pmask);  putbyte(p5 and pmask);
    waitbyte(cmd_set_path);
    setdata;
  end;

procedure path { pathid: paths; status: boolean };

  {Set specified path status on/off (true/false)}

  var n: integer;

  begin
    n := currentmfc;
    if status then
      pstat[n,pathid] := $1f
    else
      pstat[n,pathid] := 0;
    setpath(pstat[n,midiin_system], pstat[n,system_midiout],
            pstat[n,midiin_sp], pstat[n,system_sp], pstat[n,midiin_midiout]);
  end;

procedure getstatus;

  {Get node and path status of current MFC - store into nstat, pstat}

  var
    i, x: integer;
    p: paths;

  begin
    if mdebug then write(' getstatus');
    setcmd;
    putbyte(cmd_getnodestatus);
    waitbyte(cmd_getnodestatus);
    for i := 0 to 7 do
      begin
        getbyte(x);
        if i <= 5 then
          nstat[currentmfc,i] := x and 127;
      end;

    putbyte(cmd_getpathstatus);
    waitbyte(cmd_getpathstatus);
    for p := midiin_system to midiin_midiout do
      begin
        getbyte(x);
        pstat[currentmfc,p] := x and 127;
      end;
    setdata;
  end;

procedure putstatus;

  {Set node and path parameters for current MFC using nstat/pstat arrays}

  var
    i: integer;

  begin
    if mdebug then write(' putstatus');
    setcmd;
    putbyte(cmd_setnodestatus);
    for i := 0 to 5 do
      putbyte(nstat[currentmfc,i]);
    putbyte(0); putbyte(0);
    waitbyte(cmd_setnodestatus);
    setdata;

    i := currentmfc;
    setpath(pstat[i,midiin_system], pstat[i,system_midiout],
            pstat[i,midiin_sp], pstat[i,system_sp],
            pstat[i,midiin_midiout]);
  end;

procedure sendnoteon { channel, note, vel: integer } ;

  {Play note (0-127) on MIDI channel (0-15) with given velocity (0-127).
   Zero velocity = note off.}

  begin
    if mdebug then write(' noteon');
    putbyte($90+channel);
    putbyte(note);
    putbyte(vel);
  end;

procedure controlchange { channel, control, value: integer };

  {Set control (0-127) on MIDI channel (0-15) to specified value (0-127)}

  begin
    if mdebug then write(' controlchange');
    putbyte($B0+channel);
    putbyte(control);
    putbyte(value);
  end;

procedure quiet;

  {Stop all notes in progress on all MIDI channels}

  var
    i, j: integer;

  begin
    for j := 1 to 2 do
      for i := 0 to 15 do
        controlchange(i, $7b, 0);
  end;

procedure alloff { channel: integer };

  {Stop all notes in progress on a specified channel}

  var
    i: integer;

  begin
    for i := 1 to 2 do
      controlchange(channel, $7b, 0);
  end;

procedure selectvoice { channel, voiceno: integer } ;

  {Select voice (0-127) on MIDI channel (0-15)}

  begin
    if mdebug then write('selectvoice');
    putbyte($c0+channel);
    putbyte(voiceno);
  end;

procedure selectmfc { card: integer } ;

  {Select specified MFC (0/1).  Ignored if selected card not active.}

  begin
    if mdebug then write(' selectmfc:',card);
    if (card = 0) and mfc0 then
      begin
        currentmfc := 0;
        mfcbase := mfcbase0;
        node := nstat[0,stat_number];
      end
    else if (card = 1) and mfc1 then
      begin
        currentmfc := 1;
        mfcbase := mfcbase1;
        node := nstat[1,stat_number];
      end;
  end;

procedure getcardmode(var n: integer);

  {Check card mode - internal diagnostic.  Returns -1 if bad response,
   else 1 for thru mode or 0 for music mode}

  var
    t: longint;
    x: integer;

  begin
    if mdebug then write(' cardmode:');
    setcmd;
    putbyte($1d0);
    t := time;
    repeat
      getbyte(x);
    until (x > $100) or (x = $ff) or ((time-t) > 500);
    if x = $1d0 then n := 0
    else if x = $1d1 then n := 1
    else n := -1;
    if mdebug then write(n);
    setdata;
  end;


procedure initmfc;

  {Initialize interface and unit variables}

  var
    i: integer;
    p: paths;

  begin
    if mdebug then write(' initmfc');
    disableint;
    commandmode := false;
    mfc0 := istat[0] <> $ff;                 {MFC probably there }
    mfc1 := istat[1] <> $ff;                 { but still need to verify}
    mfcmode := mfc0 or mfc1;
    currentmfc := 0;
    realtime := false;                       {Do not pass real-time msgs}
    mpumode := false;
    if selectmpu then
      begin
        if mdebug then write(' MPU mode selected');
        mfc0 := false;  mfc1 := false;       {Disallow MFC use}
      end;
    if mfc0 then
      begin
        if mdebug then write(' Check unit 0');
        mfcbase := mfcbase0;                 {Port base address}
        port[mfcbase+pcr] := initpiu;        {Reset PIU}
        currentmfc := 0;
        getcardmode(i);
        if i < 0 then
          begin
            if mdebug then write('-bad response');
            mfc0 := false
          end
        else
          begin
            setpath($1f,$1f,$00,$1f,$00);      {Enable paths}
            getstatus;                         {Get node status}
            if mdebug then write('-ok');
          end
      end;
    if mfc1 then
      begin
        if mdebug then write(' Check unit 1');
        mfcbase := mfcbase1;
        port[mfcbase+pcr] := initpiu;
        currentmfc := 1;
        getcardmode(i);
        if i < 0 then
          begin
            if mdebug then write('-bad response');
            mfc1 := false
          end
        else
          begin
            setpath($1f,$1f,$00,$1f,$00);
            getstatus;
            if mdebug then write('-ok');
          end
      end;

    mfcmode := mfc0 or mfc1;
    mfctimermode := mfcmode;
    if mfc0 then selectmfc(0)
    else if mfc1 then selectmfc(1);

    if (not mfcmode) or selectmpu then
      begin                            {Check MPU}
        if mdebug then write(' check mpu-');
        setcmd;
        putbyte(cmd_reset);
        putbyte(cmd_uart);
        mpumode := lastdata = mpuack;
        if mdebug then write(mpumode);
      end;

    setdata;
    if not (mfcmode or mpumode) then
      begin
        cardpresent := false;
        writeln('MFC/MPU not active');
        delay(1000);
      end
    else
      begin
        intmode := intselected;
        cardpresent := true;
        if intmode then
          if mfc0 and mfc1 then
            intmode := false           {Single card int mode only}
          else
            begin
              if (irq < 0) then
                if mpumode then
                  irq := 2             {MPU default interrupt}
                else
                  irq := 3;            {MFC default interrupt}
              irqmask := 1 shl irq;
              intvec := 8 + irq;
              enableint;
            end;
      end;
  end;

procedure reboot;

  {Reset both MFC nodes to power-on status}

  begin
    if mdebug then write(' Reboot:');
    setcmd;
    if mfc0 then
      begin
        if mdebug then write('unit0:');
        mfcbase := mfcbase0;
        putbyte(cmd_reboot);
        delay(1000);                    {Takes a while}
        waitbyte(cmd_reboot);
      end;
    if mfc1 then
      begin
        if mdebug then write('unit1:');
        mfcbase := mfcbase1;
        putbyte(cmd_reboot);
        delay(1000);
        waitbyte(cmd_reboot);
      end;
    setdata;
    if mdebug then write('done');
    initmfc;
  end;

function checknode: boolean;

  {Check whether current node will answer requests and verify identification
   string.  Returns true if ok.}

  var
    x: integer;
    ok: boolean;

  begin
    if mdebug then write(' checknode:',node,':');
    nodedump(4, 0);
    waitbyte($f0);
    ok := lastdata > 0;
    if ok then waitbyte($75);
    ok := lastdata > 0;
    if ok then waitbyte(node);
    ok := lastdata >= 0;
    if ok then
      begin
        getbyte(x); getbyte(x); getbyte(x);
        getpacket(b_packet);
        getbyte(x);
        ok := (not packeterror) and
              (packetbuffer[1] = ord('Y')) and
              (packetbuffer[2] = ord('A'));
      end;
    checknode := ok;
    if mdebug then write(ok);
  end;

procedure testcard;

  {Diagnostic test for acceptable MFC behavior - intended for stand-alone use}

  var
    i, j, x: integer;
    stat, stat1, stat2: byte;
    p: paths;
    err: boolean;
    t: longint;

  begin
    disableint;
    if istat[0] <> $ff then
      begin mfcbase := mfcbase0;  currentmfc := 0; end
    else if istat[1] <> $ff then
      begin mfcbase := mfcbase1;  currentmfc := 1; end;
    mfcmode := true;
    writeln('Testing...');
    writeln;
    setcmd;
    err := false;
    {Test txrdy working}
    t := time;
    stat1 := inport(mfcbase+piu2);
    repeat
      stat := inport(mfcbase+piu2);
      timeoutset := abs(time-t) > 1000;
    until ((stat and txrdy) <> 0) or timeoutset;
    stat2 := stat;
    if timeoutset then
      begin
        err := true;
        writeln('PIU tx status error, stat1=',hex(stat1),
                ' stat2=',hex(stat2));
      end
    else     {Try writing something (get path status)}
      begin
        writeln('PIU tx status ok');
        port[mfcbase+piu1] := byte(cmd_getpathstatus);
        t := time;
        stat1 := inport(mfcbase+piu2);
        repeat
          stat := inport(mfcbase+piu2);
          timeoutset := abs(time-t) > 1000;
        until ((stat and txrdy) <> 0) or timeoutset;
        stat2 := stat;
        if timeoutset then
          begin
            err := true;
            writeln('PIU write error, stat1=',hex(stat1),
                    ' stat2=',hex(stat2));
          end
        else    {Try reading path status}
          begin
            writeln('PIU write ok');
            t := time;
            stat1 := inport(mfcbase+piu2);
            repeat
              stat := inport(mfcbase+piu2);
              timeoutset := abs(time-t) > 1000;
            until ((stat and rxrdy) <> 0) or timeoutset;
            stat2 := stat;
            if timeoutset then
              begin
                err := true;
                writeln('PIU read error, stat1=',hex(stat1),
                        ' stat2=',hex(stat2));
              end
            else     {No timeout, get status}
              begin
                waitbyte(cmd_getpathstatus);
                if lastdata = cmd_getpathstatus then
                  begin
                    writeln('PIU read ok');
                    for i := 1 to 5 do getbyte(x);
                  end
                else     {Couldn't read status}
                  begin
                    err := true;
                    writeln('Status command not confirmed');
                  end;
              end;
          end;
      end;
    setdata;
    if not err then for i := 0 to 1 do     {Display node/path status}
      if istat[i] <> $ff then
        begin
          write('Card ',i,' status: ');
          for j := 0 to 5 do write(hex(nstat[i,j]):4);
          writeln;
          write(' ':15);
          for p := midiin_system to midiin_midiout do
            write(hex(pstat[i,p]):4);
          if not err then
            begin
              err := not checknode;
              if err then
                writeln('Card did not respond with proper id');
            end;
        end;
    writeln;
    writeln;
    if err then
      writeln('*** Could not properly access Music Feature Card')
    else
      writeln('Music Feature Card appears to be functioning correctly');
    setdata;
  end;

procedure settimer { n: integer };

  {Set countdown timer to n 1.024 millisecond ticks.}

  var
    low, high: integer;

  begin
    if mfctimermode then
      begin
        timeoutset := false;
        timercheck := 10000;               {Safety check for inoperative timer}
        if n > 0 then
          begin
            n := n - 1;
            timerhigh := 255 - (n div 256);     {Zero point for timeout}
            timerlow := 255 - (n and 255);
            port[mfcbase+tcwr] := timer1 + lsb_msb + mode2;
            port[mfcbase+cntr1] := 255;         {Set both high and low to max}
            port[mfcbase+cntr1] := 255;
            port[mfcbase+tcwr] := timer2 + lsb_msb + mode3;
            port[mfcbase+cntr2] := 0;           {Tick duration for timer 2}
            port[mfcbase+cntr2] := 8;           { = 1.024 ms}
            port[mfcbase+tcwr] := timer2 + latch;   {Start timer}
            startlow := inport(mfcbase+cntr1);        {Read starting values}
            starthigh := inport(mfcbase+cntr1);
          end;
      end
    else {DOS timing - milliseconds}
      begin
        timeoutset := false;
        starttime := time;
        checktime := n;
      end;
  end;

function timeout: boolean;

  {Test if counter has timed out}

  var
    low, high: integer;
    started: boolean;

  begin
    if mfctimermode then
      begin
        low := inport(mfcbase+cntr1);        {Read current values}
        high := inport(mfcbase+cntr1);
        started := (starthigh <> high) or
                   (startlow <> low);      {Silly thing moving yet?}
        timercheck := timercheck - 1;
        if (timercheck < 0) and not started then
          timeoutset := true
        else
          timeoutset := started and ((low <= timerlow) and
                                    (high <= timerhigh));
      end
    else
      timeoutset :=  abs(time-starttime) > checktime;
    timeout := timeoutset;
end;


{$f+}
procedure mfcexit;

  {Clean up and exit}

  begin
    exitproc := exitsave;
    if intinstalled then
      begin
        disableint;
        setintvec(intvec, savevec);
      end;
    if mpumode then
      begin
        setcmd;
        putbyte(cmd_reset);
      end;
  end;
{$f-}

procedure params;

  { Get relevant command line parameters for calling program:

        -r   Select MPU if MFC is also present.
        -p   Set MPU port address (ex: -p330).
        -i   Select interrupt mode (add number from 0-7 to pick IRQ).
        -t   Set timeout in milliseconds (ex: -t500).
        -d   Set debug mode.
   }

  var
    i, j, n: integer;
    s: string[20];
    s2: string[2];
    s1: string[1];

  begin
    for i := 1 to paramcount do
      begin
        s := paramstr(i);
        s2 := copy(s,1,2);
        s := copy(s,3,10);
        if s2 = '-r' then              {Force use of MPU}
          selectmpu := true
        else if s2 = '-p' then         {MPU i/o port address}
          begin
            n := 0;
            for j := 1 to length(s) do
              begin
                s1 := upcase(s[j]);
                n := n * 16 + pos(s1, hexdigits) - 1;
              end;
            if n >= 0 then mpubase := n;
          end
        else if s2 = '-t' then         {I/O time out value in milliseconds}
          val(s,iolimit,n)
        else if s2 = '-i' then         {Interrupt mode}
          begin
            if s = '' then
              intselected := true
            else
              begin
                n := byte(s[1]) and 7;
                if (n >= 0) and (n <= 7) then
                  irq := n
              end;
          end
        else if s2 = '-d' then         {Debug mode}
          begin
            mdebug := true;
            assign(output,'');
            rewrite(output);
          end;
      end;
  end;

{ Unit initialization }

  begin
    exitsave := exitproc;              {Save exit address}
    exitproc := @mfcexit;
    istat[0] := inport(mfcbase0+piu2);   {Save MFC interrupt status}
    istat[1] := inport(mfcbase1+piu2);
    timeoutset := false;
    mpubase := mpubase0;               {Default MPU port addr}
    selectmpu := false;
    iolimit := iodefault;
    mdebug := false;
    midisent := false;
    waitack := true;
    intmode := false;                  {** Set true to force interrupt mode}
    intselected := false;
    intenabled := false;
    intinstalled := false;
    irq := -1;
    tcrbyte := 0;
    mfctimermode := false;
    params;                            {Check command line parameters}
    initmfc;                           {Set up everything else}
    counter := 0;
  end.

